/**
 * 	\file		errore.h
 *	\author		Tranchida Giulio, No Matricola 241732\n
 *				Si dichiara che il contenuto di questo file e',
 * 				in ogni sua parte, opera originale dell'autore.\n\n
 *
 * This program is free software; you can redistribuite it and/or modify
 * it under the terms of the GNU/General Pubblic License as published the
 * Free software Foundation; either version 2 of the License, or
 * (at your opinion) any later version.
 * */
#ifndef ERRORE_H_
#define ERRORE_H_

#include <errno.h>
#include <stdio.h>
#include <stdarg.h>

/**
 * Questa funzione variadica e' a solo scopo di debugging del codice.\n
 * La funzione si occupa di stampare a monitor diverse informazioni
 * in merito ad un errore che si e' verificato. Il comportamento della
 * funzione stessa e' diversificato in base alla varibile _DEBUG_MODE, 
 * in particolare, se questa e' definita, la funzione stampera' a monitor,
 * sullo standard error:
 * 	- 	Un messaggio che informa in queale file e in quel linea di esso
 * 		si e' verificato l'errore,
 * 	- 	un messagio di errore personalizzato a cui segue il messaggio
 * 		prodotto dalla perror,
 * 	-	il numero di errore ritornato da errno.
 * 	-	vari ed eventuali altre informazioni
 * 
 *  \param *file Il nome del file dove si e' verificato l'errore
 *  \param line Il numero di riga dove si e' verificato l'errore
 *  \param *format Un messaggio che verra' stampato a monitor
 * 
 *  \return errno  
 * */
int sys_err(const char *file, int line, const char *format, ...);

/**
 * Questa funzione variadica e' a solo scopo di debugging del codice.\n
 * La funzione si occupa di stampare a monitor diverse informazioni
 * in merito ad un errore che si e' verificato. Il comportamento della
 * funzione stessa e' diversificato in base alla varibile _DEBUG_MODE,
 * in particolare, se questa e' definita, la funzione stampera' a monitor,
 * sullo standard error:
 * 	- 	Un messaggio che informa in queale file e in quel linea di esso
 * 		si e' verificato l'errore,
 * 	- 	un messagio di errore personalizzato,
 * 	-	il numero di errore ritornato da errno,
 * 	-	vari ed eventuali altre informazioni.
 *
 *  \param *file Il nome del file dove si e' verificato l'errore
 *  \param line Il numero di riga dove si e' verificato l'errore
 *  \param *format Un messaggio che verra' stampato a monitor
 *
 *  \return errno
 * */
void gen_err(const char *file, int line, const char *format, ...);

#endif /*ERRORE_H_*/
